#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file DroneFlightDirector.py
 @brief Controlling the flying of drones.
 @date $Date$

 @author East Japan Accounting Center Co., Ltd

 BSD 3clause "New" or "Revised" License

"""
import sys
import math

# Import RTM module
import RTC
import OpenRTM_aist

import target_gps
import drone_distance_cal

# Setup logger
from logging import getLogger, StreamHandler
from logging import DEBUG as LOG_LV

sys.path.append(".")

logger = getLogger(__name__)
handler = StreamHandler()
logger.setLevel(LOG_LV)
handler.setLevel(LOG_LV)
logger.addHandler(handler)

# constants
BUTTON_01_MASK = 0x0001
BUTTON_02_MASK = 0x0002
BUTTON_03_MASK = 0x0004
BUTTON_04_MASK = 0x0008
BUTTON_05_MASK = 0x0010
BUTTON_06_MASK = 0x0020
BUTTON_07_MASK = 0x0040
BUTTON_08_MASK = 0x0080
BUTTON_09_MASK = 0x0100
BUTTON_10_MASK = 0x0200
BUTTON_11_MASK = 0x0400
BUTTON_12_MASK = 0x0800
BUTTON_13_MASK = 0x1000
BUTTON_14_MASK = 0x2000

INPORT_NUM = 3
MOVE_SPEED_MAX = 10.0
MOVE_SPEED_MIN = 1.0

MOVE_UPDATE_COEF_ABSOLUTE_NS = 1.0
MOVE_UPDATE_COEF_ABSOLUTE_EW = 0.65
MOVE_UPDATE_COEF_RELATIVE = 0.55

ALLOWABLE_ERROR_POS_CMP = 5.0
ALLOWABLE_ERROR_HEIGHT = 0.0

NO_MOVE_MODE = 0
TRANS_MODE = 1
TURN_MODE = 2

TURN_UPDATE_DIST = 3.0
ONLY_ONE_TURN_SPEED = 45

DRONE_SPEED_RANGE_TO_START_TURN = 0.1

IN_HOME_AREA = 2.0

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
droneflightdirector_spec = ["implementation_id", "DroneFlightDirector", 
         "type_name",         "DroneFlightDirector", 
         "description",       "Controlling the flying of drones.", 
         "version",           "1.0.0", 
         "vendor",            "EAC", 
         "category",          "Drone", 
         "activity_type",     "STATIC", 
         "max_instance",      "1", 
         "language",          "Python", 
         "lang_type",         "SCRIPT",
         "conf.default.AltCoefficient", "1",
         "conf.default.LatLonCoefficient", "10",
         "conf.default.LatDiff10m", "0.000137",
         "conf.default.LonDiff10m", "0.000108",
         "conf.default.NoseLatDiff10m", "9.01323",
         "conf.default.NoseLonDiff10m", "8.9831566",
         "conf.default.MoveSpeed", "3.0",
         "conf.default.MinAlt", "2.0",
         "conf.default.MaxAlt", "15.0",
         "conf.default.DroneNum", "3",
         "conf.default.Drone1TakeoffAlt", "5.0",
         "conf.default.Drone2TakeoffAlt", "5.0",
         "conf.default.Drone3TakeoffAlt", "5.0",
         "conf.__widget__.AltCoefficient", "spin",
         "conf.__widget__.LatLonCoefficient", "spin",
         "conf.__widget__.LatDiff10m", "spin",
         "conf.__widget__.LonDiff10m", "spin",
         "conf.__widget__.NoseLatDiff10m", "spin",
         "conf.__widget__.NoseLonDiff10m", "spin",
         "conf.__widget__.MoveSpeed", "spin",
         "conf.__widget__.MinAlt", "spin",
         "conf.__widget__.MaxAlt", "spin",
         "conf.__widget__.DroneNum", "spin",
         "conf.__widget__.Drone1TakeoffAlt", "spin",
         "conf.__widget__.Drone2TakeoffAlt", "spin",
         "conf.__widget__.Drone3TakeoffAlt", "spin",
         "conf.__constraints__.AltCoefficient", "1<=x<=10",
         "conf.__constraints__.LatLonCoefficient", "2<=x<=10",
         "conf.__constraints__.LatDiff10m", "0.000000<x<0.100000",
         "conf.__constraints__.LonDiff10m", "0.000000<x<0.100000",
         "conf.__constraints__.NoseLatDiff10m", "0.000000<=x<=10.000000",
         "conf.__constraints__.NoseLonDiff10m", "0.0000000<=x<=10.0000000",
         "conf.__constraints__.MoveSpeed", "1.0<=x<=10.0",
         "conf.__constraints__.MinAlt", "1.0<=x<=100.0",
         "conf.__constraints__.MaxAlt", "10.0<=x<=100.0",
         "conf.__constraints__.DroneNum", "1<=x<=3",
         "conf.__constraints__.Drone1TakeoffAlt", "x>=1.0",
         "conf.__constraints__.Drone2TakeoffAlt", "x>=1.0",
         "conf.__constraints__.Drone3TakeoffAlt", "x>=1.0",
         ""]
# </rtc-template>

##
# @class DroneFlightDirector
# @brief Controlling the flying of drones.
# 
# Refer to the "機能仕様書_DroneFlightDirector" for details.
# 
# Refer to the "機能仕様書_DroneFlightDirector" for details.
# 
# Refer to the "機能仕様書_DroneFlightDirector" for details.
# 
# Refer to the "機能仕様書_DroneFlightDirector" for details.
# 
# 
class DroneFlightDirector(OpenRTM_aist.DataFlowComponentBase):
    
    ##
    # @brief constructor
    # @param manager Maneger Object
    # 
    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_ButtonIn = RTC.TimedULong(RTC.Time(0,0),0)
        """
        PS4コントローラのボタン入力状態
         - Type: RTC::TimedULong
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._ButtonInIn = OpenRTM_aist.InPort("ButtonIn", self._d_ButtonIn)
        self._d_AnalogIn = RTC.TimedDoubleSeq(RTC.Time(0,0),[])
        """
        PS4コントローラのアナログ入力状態
         - Type: RTC::TimedDoubleSeq
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._AnalogInIn = OpenRTM_aist.InPort("AnalogIn", self._d_AnalogIn)
        self._d_Home1GPSIn = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                         0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ホーム（基地局）座標　（１号機）
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._Home1GPSInIn = OpenRTM_aist.InPort("Home1GPSIn", self._d_Home1GPSIn)
        self._d_Drone1GPSIn = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                         0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ドローン現在地座標　（１号機）
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._Drone1GPSInIn = OpenRTM_aist.InPort("Drone1GPSIn", self._d_Drone1GPSIn)
        self._d_inAccel1 = RTC.TimedFloatSeq(RTC.Time(0,0),[])
        """
        加速度　（１号機）
         - Type: RTC::TimedFloatSeq
         - Semantics: data[0]:x
                      data[1]:y
                      data[2]:z
         - Unit: m/sec
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._inAccel1In = OpenRTM_aist.InPort("inAccel1", self._d_inAccel1)
        self._d_inMag1 = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        磁気コンパス　（１号機）
         - Type: RTC::TimedFloat
         - Unit: deg
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._inMag1In = OpenRTM_aist.InPort("inMag1", self._d_inMag1)
        self._d_inState1 = RTC.TimedString(RTC.Time(0,0),0)
        """
        ドローンI/Oの内部ステータス　（１号機）
         - Type: RTC::TimedString
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._inState1In = OpenRTM_aist.InPort("inState1", self._d_inState1)
        self._d_Home2GPSIn = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                         0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ホーム（基地局）座標　（２号機）
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._Home2GPSInIn = OpenRTM_aist.InPort("Home2GPSIn", self._d_Home2GPSIn)
        self._d_Drone2GPSIn = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                         0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ドローン現在地座標　（２号機）
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._Drone2GPSInIn = OpenRTM_aist.InPort("Drone2GPSIn", self._d_Drone2GPSIn)
        self._d_inAccel2 = RTC.TimedFloatSeq(RTC.Time(0,0),[])
        """
        加速度　（２号機）
         - Type: RTC::TimedFloatSeq
         - Semantics: data[0]:x
                      data[1]:y
                      data[2]:z
         - Unit: m/sec
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._inAccel2In = OpenRTM_aist.InPort("inAccel2", self._d_inAccel2)
        self._d_inMag2 = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        磁気コンパス　（２号機）
         - Type: RTC::TimedFloat
         - Unit: deg
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._inMag2In = OpenRTM_aist.InPort("inMag2", self._d_inMag2)
        self._d_inState2 = RTC.TimedString(RTC.Time(0,0),0)
        """
        ドローンI/Oの内部ステータス　（２号機）
         - Type: RTC::TimedString
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._inState2In = OpenRTM_aist.InPort("inState2", self._d_inState2)
        self._d_Home3GPSIn = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                         0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ホーム（基地局）座標　（３号機）
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._Home3GPSInIn = OpenRTM_aist.InPort("Home3GPSIn", self._d_Home3GPSIn)
        self._d_Drone3GPSIn = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                         0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ドローン現在地座標　（３号機）
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._Drone3GPSInIn = OpenRTM_aist.InPort("Drone3GPSIn", self._d_Drone3GPSIn)
        self._d_inAccel3 = RTC.TimedFloatSeq(RTC.Time(0,0),[])
        """
        加速度　（３号機）
         - Type: RTC::TimedFloatSeq
         - Semantics: data[0]:x
                      data[1]:y
                      data[2]:z
         - Unit: m/sec
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._inAccel3In = OpenRTM_aist.InPort("inAccel3", self._d_inAccel3)
        self._d_inMag3 = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        磁気コンパス　（３号機）
         - Type: RTC::TimedFloat
         - Unit: deg
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._inMag3In = OpenRTM_aist.InPort("inMag3", self._d_inMag3)
        self._d_inState3 = RTC.TimedString(RTC.Time(0,0),0)
        """
        ドローンI/Oの内部ステータス　（３号機）
         - Type: RTC::TimedString
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._inState3In = OpenRTM_aist.InPort("inState3", self._d_inState3)
        self._d_Target1GPSOut = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                         0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ターゲット座標　（１号機）
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._Target1GPSOutOut = OpenRTM_aist.OutPort("Target1GPSOut", self._d_Target1GPSOut)
        self._d_outTarget1Speed = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        移動速度　（１号機）
         - Type: RTC::TimedFloat
         - Unit: m/sec
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._outTarget1SpeedOut = OpenRTM_aist.OutPort("outTarget1Speed", self._d_outTarget1Speed)
        self._d_outTarget1Yaw = RTC.TimedShortSeq(RTC.Time(0,0),[])
        """
        旋回方向及び移動量　（１号機）
         - Type: RTC::TimedShortSeq
         - Semantics: data[0]:旋回方向(1:右　-1:左　0:旋回無し)
                      data[1]:ターゲット絶対角(単位:deg)
         - Unit: deg
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._outTarget1YawOut = OpenRTM_aist.OutPort("outTarget1Yaw", self._d_outTarget1Yaw)
        self._d_Target2GPSOut = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                         0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ターゲット座標　（２号機）
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._Target2GPSOutOut = OpenRTM_aist.OutPort("Target2GPSOut", self._d_Target2GPSOut)
        self._d_outTarget2Speed = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        移動速度　（２号機）
         - Type: RTC::TimedFloat
         - Unit: m/sec
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._outTarget2SpeedOut = OpenRTM_aist.OutPort("outTarget2Speed", self._d_outTarget2Speed)
        self._d_outTarget2Yaw = RTC.TimedShortSeq(RTC.Time(0,0),[])
        """
        旋回方向及び移動量　（２号機）
         - Type: RTC::TimedShortSeq
         - Semantics: data[0]:旋回方向(1:右　-1:左　0:旋回無し)
                      data[1]:ターゲット絶対角(単位:deg)
         - Unit: deg
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._outTarget2YawOut = OpenRTM_aist.OutPort("outTarget2Yaw", self._d_outTarget2Yaw)
        self._d_Target3GPSOut = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                         0.0, 0, RTC.GPS_FIX_NORMAL)
        """
        ターゲット座標　（３号機）
         - Type: RTC::GPSData
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._Target3GPSOutOut = OpenRTM_aist.OutPort("Target3GPSOut", self._d_Target3GPSOut)
        self._d_outTarget3Speed = RTC.TimedFloat(RTC.Time(0,0),0)
        """
        移動速度　（３号機）
         - Type: RTC::TimedFloat
         - Unit: m/sec
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._outTarget3SpeedOut = OpenRTM_aist.OutPort("outTarget3Speed", self._d_outTarget3Speed)
        self._d_outTarget3Yaw = RTC.TimedShortSeq(RTC.Time(0,0),[])
        """
        旋回方向及び移動量　（３号機）
         - Type: RTC::TimedShortSeq
         - Semantics: data[0]:旋回方向(1:右　-1:左　0:旋回無し)
                      data[1]:ターゲット絶対角(単位:deg)
         - Unit: deg
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._outTarget3YawOut = OpenRTM_aist.OutPort("outTarget3Yaw", self._d_outTarget3Yaw)
        self._d_outMoveMode = RTC.TimedString(RTC.Time(0,0),0)
        """
        移動モード（Absolute or Relative）
         - Type: RTC::TimedString
         - Frequency: コンポーネント実行周期と同様
         - Operation Cycle: コンポーネント実行周期と同様
        """
        self._outMoveModeOut = OpenRTM_aist.OutPort("outMoveMode", self._d_outMoveMode)


        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">
        """
        移動ターゲット高度算出係数
         - Name: AltCoefficient AltCoefficient
         - DefaultValue: 1
         - Range: 1<=x<=10
        """
        self._AltCoefficient = [1]
        """
        移動ターゲット経緯度算出係数
         - Name: LatLonCoefficient LatLonCoefficient
         - DefaultValue: 10
         - Range: 2<=x<=10
        """
        self._LatLonCoefficient = [10]
        """
        10mあたりの緯度差分
         - Name: LatDiff10m LatDiff10m
         - DefaultValue: 0.000137
         - Range: 0.000000<x<0.100000
        """
        self._LatDiff10m = [0.000137]
        """
        10mあたりの経度差分
         - Name: LonDiff10m LonDiff10m
         - DefaultValue: 0.000108
         - Range: 0.000000<x<0.100000
        """
        self._LonDiff10m = [0.000108]
        """
        10mあたりの機首緯度差分
         - Name: NoseLatDiff10m NoseLatDiff10m
         - DefaultValue: 9.01323
         - Range: 0.000000<=x<=10.000000
        """
        self._NoseLatDiff10m = [9.01323]
        """
        10mあたりの機首経度差分
         - Name: NoseLonDiff10m NoseLonDiff10m
         - DefaultValue: 8.9831566
         - Range: 0.0000000<=x<=10.0000000
        """
        self._NoseLonDiff10m = [8.9831566]
        """
        ドローン移動速度
         - Name: MoveSpeed MoveSpeed
         - DefaultValue: 3.0
         - Range: 1.0<=x<=10.0
        """
        self._MoveSpeed = [3.0]
        """
        ドローン最小高度
         - Name: MinAlt MinAlt
         - DefaultValue: 2.0
         - Range: 1.0<=x<=100.0
        """
        self._MinAlt = [2.0]
        """
        ドローン最大高度
        ※ドローン最大高度(MaxAlt)の設定値は、最小高度(MinAlt)の設定値以上の値とする
        。
         - Name: MaxAlt MaxAlt
         - DefaultValue: 15.0
         - Range: 10.0<=x<=100.0
        """
        self._MaxAlt = [15.0]
        """
        制御するドローン台数
         - Name: DroneNum DroneNum
         - DefaultValue: 3
         - Range: 1<=x<=3
        """
        self._DroneNum = [3]
        """
        ドローン離陸高度　（１号機）
         - Name: Drone1TakeoffAlt Drone1TakeoffAlt
         - DefaultValue: 5.0
         - Range: 1.0<=x
        """
        self._Drone1TakeoffAlt = [5.0]
        """
        ドローン離陸高度　（２号機）
         - Name: Drone2TakeoffAlt Drone2TakeoffAlt
         - DefaultValue: 5.0
         - Range: 1.0<=x
        """
        self._Drone2TakeoffAlt = [5.0]
        """
        ドローン離陸高度　（３号機）
         - Name: Drone3TakeoffAlt Drone3TakeoffAlt
         - DefaultValue: 5.0
         - Range: 1.0<=x
        """
        self._Drone3TakeoffAlt = [5.0]
        
        # </rtc-template>

        self._d_HomeGPSIn = [0] * INPORT_NUM
        self._HomeGPSInIn = [0] * INPORT_NUM
        self._d_DroneGPSIn = [0] * INPORT_NUM
        self._DroneGPSInIn = [0] * INPORT_NUM
        self._d_inAccel = [0] * INPORT_NUM
        self._inAccelIn = [0] * INPORT_NUM
        self._d_inMag = [0] * INPORT_NUM
        self._inMagIn = [0] * INPORT_NUM
        self._d_inState = [0] * INPORT_NUM
        self._inStateIn = [0] * INPORT_NUM

        self._d_TargetGPSOut = [0] * INPORT_NUM
        self._TargetGPSOutOut = [0] * INPORT_NUM
        self._d_outTargetSpeed = [0] * INPORT_NUM
        self._outTargetSpeedOut = [0] * INPORT_NUM
        self._d_outTargetYaw = [0] * INPORT_NUM
        self._outTargetYawOut = [0] * INPORT_NUM

        self._d_HomeGPSIn[0] = self._d_Home1GPSIn
        self._HomeGPSInIn[0] = self._Home1GPSInIn
        self._d_HomeGPSIn[1] = self._d_Home2GPSIn
        self._HomeGPSInIn[1] = self._Home2GPSInIn
        self._d_HomeGPSIn[2] = self._d_Home3GPSIn
        self._HomeGPSInIn[2] = self._Home3GPSInIn
        self._d_DroneGPSIn[0] = self._d_Drone1GPSIn
        self._DroneGPSInIn[0] = self._Drone1GPSInIn
        self._d_DroneGPSIn[1] = self._d_Drone2GPSIn
        self._DroneGPSInIn[1] = self._Drone2GPSInIn
        self._d_DroneGPSIn[2] = self._d_Drone3GPSIn
        self._DroneGPSInIn[2] = self._Drone3GPSInIn
        self._d_inAccel[0] = self._d_inAccel1
        self._inAccelIn[0] = self._inAccel1In
        self._d_inAccel[1] = self._d_inAccel2
        self._inAccelIn[1] = self._inAccel2In
        self._d_inAccel[2] = self._d_inAccel3
        self._inAccelIn[2] = self._inAccel3In
        self._d_inMag[0] = self._d_inMag1
        self._inMagIn[0] = self._inMag1In
        self._d_inMag[1] = self._d_inMag2
        self._inMagIn[1] = self._inMag2In
        self._d_inMag[2] = self._d_inMag3
        self._inMagIn[2] = self._inMag3In
        self._d_inState[0] = self._d_inState1
        self._inStateIn[0] = self._inState1In
        self._d_inState[1] = self._d_inState2
        self._inStateIn[1] = self._inState2In
        self._d_inState[2] = self._d_inState3
        self._inStateIn[2] = self._inState3In

        self._d_TargetGPSOut[0] = self._d_Target1GPSOut
        self._TargetGPSOutOut[0] = self._Target1GPSOutOut
        self._d_TargetGPSOut[1] = self._d_Target2GPSOut
        self._TargetGPSOutOut[1] = self._Target2GPSOutOut
        self._d_TargetGPSOut[2] = self._d_Target3GPSOut
        self._TargetGPSOutOut[2] = self._Target3GPSOutOut
        self._d_outTargetSpeed[0] = self._d_outTarget1Speed
        self._outTargetSpeedOut[0] = self._outTarget1SpeedOut
        self._d_outTargetSpeed[1] = self._d_outTarget2Speed
        self._outTargetSpeedOut[1] = self._outTarget2SpeedOut
        self._d_outTargetSpeed[2] = self._d_outTarget3Speed
        self._outTargetSpeedOut[2] = self._outTarget3SpeedOut
        self._d_outTargetYaw[0] = self._d_outTarget1Yaw
        self._outTargetYawOut[0] = self._outTarget1YawOut
        self._d_outTargetYaw[1] = self._d_outTarget2Yaw
        self._outTargetYawOut[1] = self._outTarget2YawOut
        self._d_outTargetYaw[2] = self._d_outTarget3Yaw
        self._outTargetYawOut[2] = self._outTarget3YawOut

        self._button = False
        self._analog = False
        self._home_gps = [False] * INPORT_NUM
        self._drone_gps = [False] * INPORT_NUM
        self._accel = [False] * INPORT_NUM
        self._mag = [False] * INPORT_NUM
        self._state = [False] * INPORT_NUM

        self._target_gps = target_gps.TargetGPS()

        self._leader_set_first = True
        self._rtl_num = 0

        self._drone_diff_pos = [False] * INPORT_NUM
        self._leader_diff_pos = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                            0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                            0.0, 0, RTC.GPS_FIX_NORMAL)
        num = 0
        while num < INPORT_NUM:
            self._drone_diff_pos[num] = RTC.GPSData(
                RTC.Time(0,0), RTC.GPSTime(0,0),
                0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                0.0, 0, RTC.GPS_FIX_NORMAL)
            num += 1

        self._leader_pos = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                       0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                       0.0, 0, RTC.GPS_FIX_NORMAL)
        self._leader_target = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                          0.0, 0, RTC.GPS_FIX_NORMAL)
        self._leader_next_pos = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                            0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                            0.0, 0, RTC.GPS_FIX_NORMAL)

        self._avg_drone_pos = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                          0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                          0.0, 0, RTC.GPS_FIX_NORMAL)

        self._target = [None] * INPORT_NUM
        num = 0
        while num < INPORT_NUM:
            self._target[num] = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                            0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                            0.0, 0, RTC.GPS_FIX_NORMAL)
            num += 1

        self._mode = NO_MOVE_MODE

        self._last_front_back = 0
        self._last_right_left = 0
        self._last_up_down = 0
        self._last_speedparam = 0

        self._avg_mag = 0
        self._turn_angle = 0
        self._move_mag = 0

        self._Drone_Num = self._DroneNum[0]
        self._SpeedParam = self._MoveSpeed[0]

        self._takeoff_alt = [None] * INPORT_NUM
        self._takeoff_alt[0] = self._Drone1TakeoffAlt[0]
        self._takeoff_alt[1] = self._Drone2TakeoffAlt[0]
        self._takeoff_alt[2] = self._Drone3TakeoffAlt[0]

        # connection check for data port ButtonInIn
        func = lambda: self._cb_connect_status_change("ButtonIn", True)
        ob = ConnectionCallback(func)
        self._ButtonInIn.setOnConnected(ob)

        func = lambda: self._cb_connect_status_change("ButtonIn", False)
        ob = ConnectionCallback(func)
        self._ButtonInIn.setOnDisconnected(ob)

        # connection check for data port AnalogInIn
        func = lambda: self._cb_connect_status_change("AnalogIn", True)
        ob = ConnectionCallback(func)
        self._AnalogInIn.setOnConnected(ob)

        func = lambda: self._cb_connect_status_change("AnalogIn", False)
        ob = ConnectionCallback(func)
        self._AnalogInIn.setOnDisconnected(ob)

    ##
    # RTC初期化
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry() 
    # 
    # @return RTC::ReturnCode_t
    # 
    #
    def onInitialize(self):
        # Bind variables and configuration variable
        self.bindParameter("AltCoefficient", self._AltCoefficient, "1")
        self.bindParameter("LatLonCoefficient", self._LatLonCoefficient, "10")
        self.bindParameter("LatDiff10m", self._LatDiff10m, "0.000137")
        self.bindParameter("LonDiff10m", self._LonDiff10m, "0.000108")
        self.bindParameter("NoseLatDiff10m", self._NoseLatDiff10m, "9.01323")
        self.bindParameter("NoseLonDiff10m", self._NoseLonDiff10m, "8.9831566")
        self.bindParameter("MoveSpeed", self._MoveSpeed, "3.0")
        self.bindParameter("MinAlt", self._MinAlt, "2.0")
        self.bindParameter("MaxAlt", self._MaxAlt, "15.0")
        self.bindParameter("DroneNum", self._DroneNum, "3")
        self.bindParameter("Drone1TakeoffAlt", self._Drone1TakeoffAlt, "5.0")
        self.bindParameter("Drone2TakeoffAlt", self._Drone2TakeoffAlt, "5.0")
        self.bindParameter("Drone3TakeoffAlt", self._Drone3TakeoffAlt, "5.0")
        
        # Set InPort buffers
        self.addInPort("ButtonIn",self._ButtonInIn)
        self.addInPort("AnalogIn",self._AnalogInIn)
        self.addInPort("Home1GPSIn",self._Home1GPSInIn)
        self.addInPort("Drone1GPSIn",self._Drone1GPSInIn)
        self.addInPort("inAccel1",self._inAccel1In)
        self.addInPort("inMag1",self._inMag1In)
        self.addInPort("inState1",self._inState1In)
        self.addInPort("Home2GPSIn",self._Home2GPSInIn)
        self.addInPort("Drone2GPSIn",self._Drone2GPSInIn)
        self.addInPort("inAccel2",self._inAccel2In)
        self.addInPort("inMag2",self._inMag2In)
        self.addInPort("inState2",self._inState2In)
        self.addInPort("Home3GPSIn",self._Home3GPSInIn)
        self.addInPort("Drone3GPSIn",self._Drone3GPSInIn)
        self.addInPort("inAccel3",self._inAccel3In)
        self.addInPort("inMag3",self._inMag3In)
        self.addInPort("inState3",self._inState3In)
        
        # Set OutPort buffers
        self.addOutPort("Target1GPSOut",self._Target1GPSOutOut)
        self.addOutPort("outTarget1Speed",self._outTarget1SpeedOut)
        self.addOutPort("outTarget1Yaw",self._outTarget1YawOut)
        self.addOutPort("Target2GPSOut",self._Target2GPSOutOut)
        self.addOutPort("outTarget2Speed",self._outTarget2SpeedOut)
        self.addOutPort("outTarget2Yaw",self._outTarget2YawOut)
        self.addOutPort("Target3GPSOut",self._Target3GPSOutOut)
        self.addOutPort("outTarget3Speed",self._outTarget3SpeedOut)
        self.addOutPort("outTarget3Yaw",self._outTarget3YawOut)
        self.addOutPort("outMoveMode",self._outMoveModeOut)
        
        # Set service provider to Ports
        
        # Set service consumers to Ports
        
        # Set CORBA Service Ports
        
        return RTC.RTC_OK
    
    #    ##
    #    # 
    #    # The finalize action (on ALIVE->END transition)
    #    # formaer rtc_exiting_entry()
    #    # 
    #    # @return RTC::ReturnCode_t
    #
    #    # 
    #def onFinalize(self):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The startup action when ExecutionContext startup
    #    # former rtc_starting_entry()
    #    # 
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onStartup(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The shutdown action when ExecutionContext stop
    #    # former rtc_stopping_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onShutdown(self, ec_id):
    #
    #    return RTC.RTC_OK
    
        ##
        # 変数初期化
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        # 
        # @return RTC::ReturnCode_t
        #
        #
    def onActivated(self, ec_id):
        self._Drone_Num = self._DroneNum[0]
        self._SpeedParam = self._MoveSpeed[0]
        self._leader_set_first = True
        self._rtl_num = 0
        if self._is_landing() == False:
            self._landing_flag = False

        num = 0
        while num < self._Drone_Num:
            self._d_inState[num].data = "None"
            self._target[num] = RTC.GPSData(RTC.Time(0,0), RTC.GPSTime(0,0),
                                            0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                            0.0, 0, RTC.GPS_FIX_NORMAL)
            num += 1

        self._execute_rate = self.getExecutionRate(ec_id)

        init_movemode = 1
        self._target_gps.set_controller_mode(init_movemode)

        if self._takeoff_range_check() == False:
            return RTC.RTC_ERROR

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The deactivated action (Active state exit action)
    #    # former rtc_active_exit()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onDeactivated(self, ec_id):
    #
    #    return RTC.RTC_OK
    
        ##
        # PS4キー入力、ドローンステータス取得、およびコントロールデータ送信の周期処理。
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):
        self._read_inport()

        self._move_speed()
        self._set_outport()

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The aborting action when main logic error occurred.
    #    # former rtc_aborting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onAborting(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The error action in ERROR state
    #    # former rtc_error_do()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onError(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The reset action that is invoked resetting
    #    # This is same but different the former rtc_init_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onReset(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The state update action that is invoked after onExecute() action
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #

    #    #
    #def onStateUpdate(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The action that is invoked when execution context's rate is changed
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onRateChanged(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    def _read_inport(self):
        self._home_gps = [False] * self._Drone_Num
        self._drone_gps = [False] * self._Drone_Num
        self._accel = [False] * self._Drone_Num
        self._mag = [False] * self._Drone_Num
        self._state = [False] * self._Drone_Num

        # read_button
        self._d_ButtonIn = RTC.TimedULong(RTC.Time(0, 0), 0)

        if self._button:
            self._d_ButtonIn = self._ButtonInIn.read()
            logger.debug("Button = 0b" + format(self._d_ButtonIn.data, '016b'))
        else:
            logger.error("ERROR: ButtonIn is NOT connected")

        # read_analog
        self._d_AnalogIn = RTC.TimedDoubleSeq(RTC.Time(0, 0), [])

        if self._analog:
            self._d_AnalogIn = self._AnalogInIn.read()
            for i in range(0, len(self._d_AnalogIn.data), 2):
                d1 = d2 = None
                try:
                    d1 = self._d_AnalogIn.data[i]
                    d2 = self._d_AnalogIn.data[i + 1]
                except IndexError:
                    pass

                if d2 is None:
                    logger.debug("Analog[%d]:  % f", i, d1)
                else:
                    logger.debug("Analog[%d,%d]: % f % f", i, i + 1, d1, d2)
        else:
            logger.error("ERROR: AnalogIn is NOT connected")

        num = 0
        while num < self._Drone_Num:
            # read_home_gps
            if self._HomeGPSInIn[num].isNew():
                self._d_HomeGPSIn[num] = self._HomeGPSInIn[num].read()
                self._home_gps[num] = True
                logger.debug("Home%dGPS:  % f % f % f",
                             (num + 1),
                             self._d_HomeGPSIn[num].latitude,
                             self._d_HomeGPSIn[num].longitude,
                             self._d_HomeGPSIn[num].altitude)

            # read_drone_gps
            if self._DroneGPSInIn[num].isNew():
                self._d_DroneGPSIn[num] = self._DroneGPSInIn[num].read()
                self._drone_gps[num] = True
                logger.debug("Drone%dGPS:  % f % f % f",
                             (num + 1),
                             self._d_DroneGPSIn[num].latitude,
                             self._d_DroneGPSIn[num].longitude,
                             self._d_DroneGPSIn[num].altitude)

            # read_accel
            if self._inAccelIn[num].isNew():
                self._d_inAccel[num] = self._inAccelIn[num].read()
                self._accel[num] = True
                logger.debug("Accel%d:  % f % f % f",
                             (num + 1),
                             self._d_inAccel[num].data[0],
                             self._d_inAccel[num].data[1],
                             self._d_inAccel[num].data[2])

            # read_mag
            if self._inMagIn[num].isNew():
                self._d_inMag[num] = self._inMagIn[num].read()
                self._mag[num] = True
                logger.debug("Mag%d:  % d",
                             (num + 1),
                             self._d_inMag[num].data)

            # read_state
            if self._inStateIn[num].isNew():
                self._d_inState[num] = self._inStateIn[num].read()
                self._state[num] = True
                logger.debug("State%d:  % s",
                             (num + 1),
                             self._d_inState[num].data)

            num += 1

    def _set_outport(self):
        if self._check_outport_exe():
            self._set_move_mode()
            self._set_target_gps()
            self._set_target_speed()
            self._set_target_yaw()

            num = 0
            while num < self._Drone_Num:
                self._TargetGPSOutOut[num].write()
                logger.debug("Target%dGPS: % f % f % f",
                            (num + 1),
                             self._d_TargetGPSOut[num].latitude,
                             self._d_TargetGPSOut[num].longitude,
                             self._d_TargetGPSOut[num].altitude)

                self._outTargetSpeedOut[num].write()
                logger.debug("Target%dSpeed: %f",
                             (num + 1),
                             self._d_outTargetSpeed[num].data)

                self._outTargetYawOut[num].write()
                logger.debug("Target%dYaw: (dir)%d, (deg)%d",
                            (num + 1),
                             self._d_outTargetYaw[num].data[0],
                             self._d_outTargetYaw[num].data[1])

                num += 1

            self._outMoveModeOut.write()
            logger.debug("MoveMode: %s",
                         self._d_outMoveMode.data)

        else:
            home = self._is_set_to_home()
            if home:
                num = 0
                while num < self._Drone_Num:
                    if self._home_gps[num]:
                        self._d_TargetGPSOut[num].latitude = self._d_HomeGPSIn[num].latitude
                        self._d_TargetGPSOut[num].longitude = self._d_HomeGPSIn[num].longitude
                        self._d_TargetGPSOut[num].altitude = self._d_HomeGPSIn[num].altitude
                        self._TargetGPSOutOut[num].write()
                        logger.debug("Target%dGPS: % f % f % f",
                                    (num + 1),
                                     self._d_TargetGPSOut[num].latitude,
                                     self._d_TargetGPSOut[num].longitude,
                                     self._d_TargetGPSOut[num].altitude)

                    num += 1

    def _check_outport_exe(self):
        ret = True
        self._all_rtl = True

        if self._landing_flag:
            ret = False

        num = 0
        while num < self._Drone_Num:
            if self._d_inState[num].data not in ["RTL","Disarm"]:
                if not self._home_gps[num] & self._drone_gps[num]:
                    logger.error("ERROR: GPS data is NOT available")
                    ret = False

                if not self._mag[num]:
                    logger.error("ERROR: MAG data is NOT available")
                    ret = False

                self._all_rtl = False
            num += 1

        if self._all_rtl:
            ret = False

        return ret

    def _is_set_to_home(self):
        home = False
        if self._all_rtl:
            if self._landing_flag:
                home = self._is_all_drone_on_home()
            else:
                home = True

        return home

    def _is_all_drone_on_home(self):
        home = True

        dist = [None] * self._Drone_Num
        num = 0
        dist_cal = drone_distance_cal.DroneMathmetics()
        while num < self._Drone_Num:
            if self._d_inState[num].data == "RTL":
                dist[num] = dist_cal.distance_math(
                    self._d_HomeGPSIn[num].longitude,
                    self._d_DroneGPSIn[num].longitude,
                    self._d_HomeGPSIn[num].latitude,
                    self._d_DroneGPSIn[num].latitude,
                    0,
                    0)

                if not (dist[num] <= IN_HOME_AREA):
                    home = False
                    break

            num += 1

        return home

    def _config_target_gps(self):
        config = {"AltCoefficient": self._AltCoefficient[0],
                  "LatLonCoefficient": self._LatLonCoefficient[0],
                  "LatDiff10m": self._LatDiff10m[0],
                  "LonDiff10m": self._LonDiff10m[0],
                  "NoseLatDiff10m": self._NoseLatDiff10m[0],
                  "NoseLonDiff10m": self._NoseLonDiff10m[0],
                  "MoveSpeed": self._MoveSpeed[0],
                  "MinAlt": self._MinAlt[0] - min(self._alt_diff),
                  "MaxAlt": self._MaxAlt[0] - max(self._alt_diff),
                  "DroneNum": self._Drone_Num
                  }

        self._target_gps.config_param(config)

    def _set_leader(self):
        self.leader_set = False
        num = 0
        count = 0
        while num < self._Drone_Num:
            if self._d_inState[num].data in ["RTL","Disarm"]:
                count += 1
            num += 1

        if self._leader_set_first:
            self._leader_set_first = False
            self.leader_set = True
        else:
            if self._rtl_num < count:
                self._rtl_num = count
                self.leader_set = True

        self._calc_leader_pos()

        if self.leader_set:
            self._set_leader_pos()
            self._drone_diff()
            self._mode = NO_MOVE_MODE
            self.leader_set = False

    def _calc_leader_pos(self):
        self._avg_drone_pos.latitude = 0.0
        self._avg_drone_pos.longitude = 0.0
        self._avg_drone_pos.altitude = 0.0
        self._avg_mag = 0
        self._avg_speed = 0

        mag_normal_result = None

        num = 0
        while num < self._Drone_Num:
            if self._d_inState[num].data not in ["RTL","Disarm"]:
                # sum of drone_pos
                self._avg_drone_pos.latitude += self._d_DroneGPSIn[num].latitude
                self._avg_drone_pos.longitude += self._d_DroneGPSIn[num].longitude
                if self._d_DroneGPSIn[num].altitude < self._MinAlt[0]:
                    self._avg_drone_pos.altitude += self._takeoff_alt[num]
                else:
                    self._avg_drone_pos.altitude += self._d_DroneGPSIn[num].altitude
                # sum of mag
                if mag_normal_result == None:
                    mag_normal_result = self._d_inMag[num].data
                else:
                    diff = self._d_inMag[num].data - self._d_inMag[last_calc_drone_num].data
                    if diff < -180:
                        diff += 360
                    elif diff > 180:
                        diff -= 360
                    mag_normal_result += diff
                self._avg_mag += mag_normal_result
                last_calc_drone_num = num

                # sum of speed
                self._avg_speed += math.sqrt((self._d_inAccel[num].data[0] * self._d_inAccel[num].data[0]) +
                                             (self._d_inAccel[num].data[1] * self._d_inAccel[num].data[1]))

            num += 1

        # average of drone_pos
        self._avg_drone_pos.latitude /= (self._Drone_Num - self._rtl_num)
        self._avg_drone_pos.longitude /= (self._Drone_Num - self._rtl_num)
        self._avg_drone_pos.altitude /= (self._Drone_Num - self._rtl_num)

        # average of mag
        self._avg_mag /= (self._Drone_Num - self._rtl_num)
        if self._avg_mag > 360:
            self._avg_mag = self._avg_mag % 360
        elif self._avg_mag < 0:
            self._avg_mag = 360 - (abs(self._avg_mag) % 360)

        # average of speed
        self._avg_speed /= (self._Drone_Num - self._rtl_num)

    def _set_leader_pos(self):
        self._leader_pos.latitude = self._avg_drone_pos.latitude
        self._leader_pos.longitude = self._avg_drone_pos.longitude
        if self.leader_set:
            self._leader_pos.altitude = self._avg_drone_pos.altitude

    def _drone_diff(self):
        self._alt_diff = [0]

        num = 0
        while num < self._Drone_Num:
            if self._d_inState[num].data not in ["RTL","Disarm"]:
                self._drone_diff_pos[num].latitude = self._d_DroneGPSIn[num].latitude - self._leader_pos.latitude
                self._drone_diff_pos[num].longitude = self._d_DroneGPSIn[num].longitude - self._leader_pos.longitude
                if self._d_DroneGPSIn[num].altitude < self._MinAlt[0]:
                    self._drone_diff_pos[num].altitude = self._takeoff_alt[num] - self._leader_pos.altitude
                else:
                    self._drone_diff_pos[num].altitude = self._d_DroneGPSIn[num].altitude - self._leader_pos.altitude

                self._alt_diff.append(self._drone_diff_pos[num].altitude)
            num += 1

    def _diff_target_angle(self, target, pos):
        angle_vec_x = math.cos(math.radians(self._avg_mag))
        angle_vec_y = math.sin(math.radians(self._avg_mag))

        target_vec_x = target.latitude - pos.latitude
        target_vec_y = target.longitude - pos.longitude

        inner_product = angle_vec_x * target_vec_x + angle_vec_y * target_vec_y
        cross_product = angle_vec_x * target_vec_y - angle_vec_y * target_vec_x

        angle_rad = math.atan2(cross_product, inner_product)
        angle_deg = math.degrees(angle_rad)

        return angle_deg

    def _set_pivot(self):
        pivot = [None] * 2
        pivot_dist = [0] * 2
        num = 0
        while num < self._Drone_Num:
            if self._d_inState[num].data not in ["RTL","Disarm"]:
                dist_cal = drone_distance_cal.DroneMathmetics()
                dist = dist_cal.distance_math(self._avg_drone_pos.longitude,
                                              self._d_DroneGPSIn[num].longitude,
                                              self._avg_drone_pos.latitude,
                                              self._d_DroneGPSIn[num].latitude,
                                              0,
                                              0)

                if pivot_dist[0] < pivot_dist[1]:
                    if pivot_dist[0] <= dist:
                        pivot_dist[0] = dist
                        pivot[0] = num
                        if pivot[1] == None:
                            pivot[1] = num
                else:
                    if pivot_dist[1] <= dist:
                        pivot_dist[1] = dist
                        pivot[1] = num
                        if pivot[0] == None:
                            pivot[0] = num
            num += 1

        angle_deg = self._diff_target_angle(self._d_DroneGPSIn[pivot[0]], self._leader_pos)
        if angle_deg >= 0:
            self._right_pivot = pivot[0]
            self._left_pivot = pivot[1]
        else:
            self._right_pivot = pivot[1]
            self._left_pivot = pivot[0]

    def _set_target_gps(self):
        self._set_leader()
        self._config_target_gps()

        takeoff_landing = self._takeoff_landing()
        if takeoff_landing == 1 or takeoff_landing == -1:
            self._leader_pos = self._target_gps. \
                get_takeoff_landing_data(
                    self._leader_pos, takeoff_landing, self._leader_pos.altitude)
            num = 0
            while num < self._Drone_Num:
                if self._d_inState[num].data not in ["RTL","Disarm"]:
                    self._target[num] = self._target_gps. \
                        get_takeoff_landing_data(
                            self._d_HomeGPSIn[num], takeoff_landing, self._takeoff_alt[num])
                num += 1

            if takeoff_landing == -1:
                self._landing_flag = True

        else:
            mv = self._moving()
            turn_dir = self._turn_dir()

            if mv.front_back != 0 or mv.right_left != 0:
                self._mode = TRANS_MODE
            elif turn_dir == 1 or turn_dir == -1:
                if self._mode != TURN_MODE:
                    if self._avg_speed < DRONE_SPEED_RANGE_TO_START_TURN:
                        self._mode = TURN_MODE

                        self._set_pivot()
                        if turn_dir == 1:
                            self._turn_pivot = self._right_pivot
                            self._turn_end = self._left_pivot
                        else:
                            self._turn_pivot = self._left_pivot
                            self._turn_end = self._right_pivot

                        self._pivot_first = self._d_DroneGPSIn[self._turn_pivot]
                        self._turn_angle_sum = 0
                        self._leader_pos_first = self._leader_pos
                        self._drone_diff_first = [0] * self._Drone_Num
                        num = 0
                        while num < self._Drone_Num:
                            if self._d_inState[num].data not in ["RTL","Disarm"]:
                                self._drone_diff_first[num] = self._drone_diff_pos[num]
                            num += 1
                    else:
                        self._mode = NO_MOVE_MODE
                else:
                    pass
            else:
                self._mode = NO_MOVE_MODE

            # translational
            self._leader_pos.altitude = self._avg_drone_pos.altitude
            if self._mode == TRANS_MODE:
                target_update = self._update_target_gps(mv.front_back, mv.right_left, mv.up_down)
                if target_update:
                    # leader target position
                    self._leader_target = self._target_gps.get_move_target_data(
                        self._leader_pos, mv.front_back, mv.right_left, mv.up_down,
                        self._SpeedParam, self._move_mag)
                    # drone target position
                    num = 0
                    while num < self._Drone_Num:
                        if self._d_inState[num].data not in ["RTL","Disarm"]:
                            self._target_gps.get_target_cal(self._target[num], self._leader_target, self._drone_diff_pos[num])
                        num += 1

                # leader next position
                dist_cal = drone_distance_cal.DroneMathmetics()
                dist_leader_target = dist_cal.distance_math(
                    self._leader_pos.longitude,
                    self._leader_target.longitude,
                    self._leader_pos.latitude,
                    self._leader_target.latitude,
                    0,
                    0)
                pos_diff_lat = self._leader_target.latitude - self._leader_pos.latitude
                pos_diff_lon = self._leader_target.longitude - self._leader_pos.longitude

                self._leader_next_pos.latitude = self._leader_pos.latitude + \
                    (pos_diff_lat * (
                        (self._avg_speed / self._execute_rate) / dist_leader_target))
                self._leader_next_pos.longitude = self._leader_pos.longitude + \
                    (pos_diff_lon * (
                        (self._avg_speed / self._execute_rate) / dist_leader_target))
                self._leader_next_pos.altitude = self._leader_target.altitude

            # turn
            elif self._mode == TURN_MODE:
                dist_cal = drone_distance_cal.DroneMathmetics()
                target_dist = dist_cal.distance_math(
                    self._d_DroneGPSIn[self._turn_end].longitude,
                    self._target[self._turn_end].longitude,
                    self._d_DroneGPSIn[self._turn_end].latitude,
                    self._target[self._turn_end].latitude,
                    0,
                    0)

                if target_dist < TURN_UPDATE_DIST:
                    # target angular velocity
                    dist_cal = drone_distance_cal.DroneMathmetics()
                    dist = dist_cal.distance_math(
                        self._d_DroneGPSIn[self._right_pivot].longitude,
                        self._d_DroneGPSIn[self._left_pivot].longitude,
                        self._d_DroneGPSIn[self._right_pivot].latitude,
                        self._d_DroneGPSIn[self._left_pivot].latitude,
                        0,
                        0)

                    if dist == 0:
                        self._turn_angle = 0
                    else:
                        self._turn_angle = (180.0 * self._SpeedParam) / (math.pi * dist)

                    self._turn_angle_sum += self._turn_angle
                    if self._turn_angle_sum > 360:
                        self._turn_angle_sum = self._turn_angle_sum % 360
                    elif self._turn_angle_sum < 0:
                        self._turn_angle_sum = 360 - (abs(self._turn_angle_sum) % 360)

                    # leader target position
                    self._leader_pos_first.altitude = self._avg_drone_pos.altitude
                    self._leader_target = self._target_gps.get_turn_target_data(
                        self._pivot_first, self._leader_pos_first,
                        turn_dir, mv.up_down, self._turn_angle_sum)
                    #
                    self._leader_pos_first.altitude = self._leader_target.altitude
                    # drone target position
                    num = 0
                    while num < self._Drone_Num:
                        if self._d_inState[num].data not in ["RTL","Disarm"]:
                            self._drone_diff_pos[num] = self._target_gps.get_turn_target_data(
                                self._leader_diff_pos, self._drone_diff_first[num],
                                turn_dir, 0, self._turn_angle_sum)
                            self._drone_diff_pos[num].altitude = self._drone_diff_first[num].altitude
                            self._target_gps.get_target_cal(
                                self._target[num], self._leader_target, self._drone_diff_pos[num])
                        num += 1

                    # leader next position
                    self._leader_next_pos = self._leader_target

            # hovering or cancel operation or home
            else:
                # leader target position
                self._set_leader_pos()
                self._leader_target = self._target_gps.get_move_target_data(
                    self._leader_pos, mv.front_back, mv.right_left, mv.up_down,
                    self._SpeedParam, self._avg_mag)
                # drone target position
                num = 0
                while num < self._Drone_Num:
                    if self._d_inState[num].data not in ["RTL","Disarm"]:
                        self._target_gps.get_target_cal(
                            self._target[num], self._leader_target, self._drone_diff_pos[num])
                    num += 1

                # leader next position
                self._leader_next_pos = self._leader_target

            self._last_front_back = mv.front_back
            self._last_right_left = mv.right_left
            self._last_up_down = mv.up_down
            self._last_speedparam = self._SpeedParam

            self._leader_pos = self._leader_next_pos

        num = 0
        while num < self._Drone_Num:
            if self._d_inState[num].data not in ["RTL","Disarm"]:
                self._d_TargetGPSOut[num].latitude = self._target[num].latitude
                self._d_TargetGPSOut[num].longitude = self._target[num].longitude
                self._d_TargetGPSOut[num].altitude = self._target[num].altitude
            else:
                self._d_TargetGPSOut[num].latitude = self._d_HomeGPSIn[num].latitude
                self._d_TargetGPSOut[num].longitude = self._d_HomeGPSIn[num].longitude
                self._d_TargetGPSOut[num].altitude = self._d_HomeGPSIn[num].altitude
            num += 1

    def _update_target_gps(self, front_back, right_left, up_down):
        target_update = True

        move_continue = self._move_continue(front_back, right_left, up_down)
        if move_continue:
            dist_cal = drone_distance_cal.DroneMathmetics()
            dist_leader_target = dist_cal.distance_math(
                self._leader_pos.longitude,
                self._leader_target.longitude,
                self._leader_pos.latitude,
                self._leader_target.latitude,
                0,
                0)

            if self._d_outMoveMode.data == "Absolute":
                if front_back:
                    update_coefficient = MOVE_UPDATE_COEF_ABSOLUTE_NS
                else:
                    update_coefficient = MOVE_UPDATE_COEF_ABSOLUTE_EW
            else:
                update_coefficient = MOVE_UPDATE_COEF_RELATIVE

            if self._last_speedparam == self._SpeedParam:
                if dist_leader_target >= (self._SpeedParam * self._LatLonCoefficient[0] * update_coefficient):
                    target_update = False
        else:
            self._move_mag = self._avg_mag

        return target_update

    def _set_target_speed(self):
        if self._mode == TRANS_MODE:
            self._trans_target_speed()
        elif self._mode == TURN_MODE:
            self._turn_target_speed()

    def _trans_target_speed(self):
        # leader next target speed
        leader_next_speed = self._SpeedParam

        # drone target speed
        dist_cal = drone_distance_cal.DroneMathmetics()
        dist_leader_target = dist_cal.distance_math(self._leader_pos.longitude,
                                                    self._leader_target.longitude,
                                                    self._leader_pos.latitude,
                                                    self._leader_target.latitude,
                                                    0,
                                                    0)

        num = 0
        while num < self._Drone_Num:
            if self._d_inState[num].data not in ["RTL","Disarm"]:
                dist_drone_target = dist_cal.distance_math(
                    self._d_DroneGPSIn[num].longitude,
                    self._d_TargetGPSOut[num].longitude,
                    self._d_DroneGPSIn[num].latitude,
                    self._d_TargetGPSOut[num].latitude,
                    0,
                    0)
                if dist_leader_target == 0:
                    drone_target_speed = leader_next_speed
                else:
                    drone_target_speed = leader_next_speed * (dist_drone_target / dist_leader_target)

                self._d_outTargetSpeed[num].data = drone_target_speed

            num += 1

    def _turn_target_speed(self):
        dist_cal = drone_distance_cal.DroneMathmetics()
        dist = dist_cal.distance_math(
            self._d_DroneGPSIn[self._right_pivot].longitude,
            self._d_DroneGPSIn[self._left_pivot].longitude,
            self._d_DroneGPSIn[self._right_pivot].latitude,
            self._d_DroneGPSIn[self._left_pivot].latitude,
            0,
            0)

        # drone target speed
        leader_real_speed = self._SpeedParam
        num = 0
        dist_cal = drone_distance_cal.DroneMathmetics()
        while num < self._Drone_Num:
            if self._d_inState[num].data not in ["RTL","Disarm"]:
                rotation_axis_dist = dist_cal.distance_math(
                    self._d_DroneGPSIn[self._turn_pivot].longitude,
                    self._d_DroneGPSIn[num].longitude,
                    self._d_DroneGPSIn[self._turn_pivot].latitude,
                    self._d_DroneGPSIn[num].latitude,
                    0,
                    0)

                if dist == 0:
                    drone_target_speed = 0
                else:
                    drone_target_speed = leader_real_speed * (rotation_axis_dist / dist)

                self._d_outTargetSpeed[num].data = drone_target_speed

            num += 1

    def _set_target_yaw(self):
        angle_deg = 0
        target_deg = 0
        target_dir = 0

        if self._mode == TRANS_MODE:
            if self._d_outMoveMode.data == "Absolute":
                angle_deg = self._diff_target_angle(self._leader_target, self._leader_pos)
                if angle_deg > 0:
                    target_dir = 1
                else:
                    target_dir = -1

        elif self._mode == TURN_MODE:
            target_dir = int(self._turn_dir())

            if self._right_pivot == self._left_pivot:
                angle_deg = ONLY_ONE_TURN_SPEED * target_dir
            else:
                angle_deg = self._diff_target_angle(self._d_TargetGPSOut[self._turn_end], self._d_DroneGPSIn[self._turn_end])

        # drone target angle
        target_deg = int(self._avg_mag + angle_deg)
        if target_deg > 360:
            target_deg = target_deg % 360
        elif target_deg < 0:
            target_deg = 360 - (abs(target_deg) % 360)

        num = 0
        while num < self._Drone_Num:
            if self._d_inState[num].data not in ["RTL","Disarm"]:
                self._d_outTargetYaw[num].data = [target_dir, target_deg]
            num += 1

    def _set_move_mode(self):
        mode = self._move_mode()

        if mode == 1:
            self._d_outMoveMode.data = "Absolute"
        else:
            self._d_outMoveMode.data = "Relative"

        self._target_gps.set_controller_mode(mode)

    def _check_drone_on_home(self):
        ret = True

        dist = [None] * self._Drone_Num
        num = 0
        dist_cal = drone_distance_cal.DroneMathmetics()
        while num < self._Drone_Num:
            if self._d_inState[num].data not in ["RTL","Disarm"]:
                dist[num] = dist_cal.distance_math(
                    self._d_HomeGPSIn[num].longitude,
                    self._d_DroneGPSIn[num].longitude,
                    self._d_HomeGPSIn[num].latitude,
                    self._d_DroneGPSIn[num].latitude,
                    0,
                    0)

                if not (dist[num] <= ALLOWABLE_ERROR_POS_CMP and
                        self._d_DroneGPSIn[num].altitude < (self._MinAlt[0] - ALLOWABLE_ERROR_HEIGHT)):
                    ret = False
                    break

            num += 1

        return ret

    def _takeoff_landing(self):
        operation = 0
        home = self._check_drone_on_home()
        if home:
            # check takeoff request
            if self._d_ButtonIn.data & BUTTON_06_MASK:
                operation = 1
            elif self._d_ButtonIn.data & BUTTON_08_MASK:
                logger.warning("WARNING: Landing operation is "
                               "NOT available "
                               "in the current drone position")
        else:
            # check landing request
            if self._d_ButtonIn.data & BUTTON_08_MASK:
                operation = -1
            elif self._d_ButtonIn.data & BUTTON_06_MASK:
                logger.warning("WARNING: Take-off operation is "
                               "NOT available in the current drone position")

        return operation

    def _move_mode(self):
        home = self._check_drone_on_home()
        mode = self._target_gps._mode

        if home:
            if self._d_ButtonIn.data & BUTTON_10_MASK:
                logger.warning("WARNING: MoveMode operation is "
                               "NOT available in the current drone position")
        else:
            mv1 = self._front_back()
            mv2 = self._right_left()
            mv3 = self._up_down()
            if mv1 == mv2 == mv3 == False:
                if self._d_ButtonIn.data & BUTTON_10_MASK:
                    if mode == 1:
                        mode = 2
                    else:
                        mode = 1

        return mode

    def _moving(self):
        mv1 = self._front_back()
        mv2 = self._right_left()
        mv3 = self._up_down()

        home = self._check_drone_on_home()

        if home:
            if mv1 or mv2 or mv3:
                mv1 = mv2 = mv3 = 0
                logger.warning("WARNING: Move operations are "
                               "NOT available in the current drone position")

        return self.MovingOperation(mv1, mv2, mv3)

    def _front_back(self):
        key = 0
        try:
            key = self._d_AnalogIn.data[0]
        except IndexError:
            pass

        operation = self._front_back_check(key)
        operation *= -1

        return operation

    def _right_left(self):
        operation = 0
        both_push = BUTTON_03_MASK | BUTTON_01_MASK
        if (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [right/left]")
        elif self._d_ButtonIn.data & BUTTON_03_MASK:
            operation = 1
        elif self._d_ButtonIn.data & BUTTON_01_MASK:
            operation = -1

        return operation

    def _up_down(self):
        operation = 0
        both_push = BUTTON_04_MASK | BUTTON_02_MASK
        if (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [up/down]")
        elif self._d_ButtonIn.data & BUTTON_04_MASK:
            operation = 1
        elif self._d_ButtonIn.data & BUTTON_02_MASK:
            operation = -1

        return operation

    def _move_continue(self, front_back, right_left, up_down):
        if (front_back != self._last_front_back) or (right_left != self._last_right_left) or (up_down != self._last_up_down):
            move_continue = False
        elif (front_back == 0) and (right_left == 0) and (up_down == 0):
            move_continue = False
        else:
            move_continue = True

        return move_continue

    def _turn_dir(self):
        home = self._check_drone_on_home()
        key = 0

        try:
            key = self._d_AnalogIn.data[1]
        except IndexError:
            pass

        if home:
            if key:
                key = 0
                logger.warning("WARNING: Turning operation is "
                               "NOT available in the current drone position")

        else:
            if self._d_ButtonIn.data & BUTTON_08_MASK:
                key = 0
            else:
                if self._d_outMoveMode.data == "Absolute":
                    mv1 = self._front_back()
                    mv2 = self._right_left()
                    if mv1 or mv2:
                        key = 0

        operation = self._turn_dir_check(key)

        return operation

    def _move_speed(self):
        home = self._check_drone_on_home()
        both_push = BUTTON_05_MASK | BUTTON_07_MASK

        if home:
            if (self._d_ButtonIn.data & BUTTON_05_MASK) or (self._d_ButtonIn.data & BUTTON_07_MASK) or (self._d_ButtonIn.data & both_push):
                logger.warning("WARNING: Speed operation is "
                               "NOT available in the current drone position")
        elif (self._d_ButtonIn.data & both_push) == both_push:
            logger.warning("WARNING: "
                           "Conflicting operations are detected [MoveSpeed up/down]")
        elif self._d_ButtonIn.data & BUTTON_05_MASK:
            self._SpeedParam += 1.0

        elif self._d_ButtonIn.data & BUTTON_07_MASK:
            self._SpeedParam -= 1.0

        self._SpeedParam = self._move_speed_check(self._SpeedParam)

    def _front_back_check(self, status):
        if status > 1:
            status = 1
        elif status < -1:
            status = -1

        return status

    def _turn_dir_check(self, dir):
        if dir > 1:
            dir = 1
        elif dir < -1:
            dir = -1

        return dir

    def _move_speed_check(self, speed):
        if speed > MOVE_SPEED_MAX:
            speed = MOVE_SPEED_MAX
        elif speed < MOVE_SPEED_MIN:
            speed = MOVE_SPEED_MIN

        return speed

    def _is_landing(self):
        landing = False
        num = 0
        while num < self._Drone_Num:
            if self._d_inState[num].data == "RTL":
                landing = True
            num += 1

        return landing

    def _takeoff_range_check(self):
        ret = True
        self._takeoff_alt = [None] * INPORT_NUM
        self._takeoff_alt[0] = self._Drone1TakeoffAlt[0]
        self._takeoff_alt[1] = self._Drone2TakeoffAlt[0]
        self._takeoff_alt[2] = self._Drone3TakeoffAlt[0]

        num = 0
        while num < self._Drone_Num:
            if (self._takeoff_alt[num] < self._MinAlt[0]) or (self._takeoff_alt[num] > self._MaxAlt[0]):
                logger.error("ERROR: Takeoff altitude is out of range")
                ret = False
                break
            num += 1

        return ret

    def _cb_connect_status_change(self, port, connect):
        if port == "ButtonIn":
            self._button = connect
        elif port == "AnalogIn":
            self._analog = connect
        else:
            logger.warning("Invalid port name: %s", port)

        if not connect:
            logger.warning("ERROR: Data port was disconnected: %s", port)

    class MovingOperation(object):

        def __init__(self, front_back, right_left, up_down):
            self.front_back = front_back
            self.right_left = right_left
            self.up_down = up_down


class ConnectionCallback(OpenRTM_aist.PortCallBack.ConnectionCallback):

    def __init__(self, func):
        self._plugin_func = func

    def __call__(self, profile):
        self._plugin_func()


def DroneFlightDirectorInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=droneflightdirector_spec)
    manager.registerFactory(profile,
                            DroneFlightDirector,
                            OpenRTM_aist.Delete)

def MyModuleInit(manager):
    DroneFlightDirectorInit(manager)

    # Create a component
    comp = manager.createComponent("DroneFlightDirector")

def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager()

if __name__ == "__main__":
    main()

